@@ -69,11 +69,12 @@ struct _CcAvatarChooser {
 G_DEFINE_TYPE (CcAvatarChooser, cc_avatar_chooser, GTK_TYPE_POPOVER)
 
 static void
-crop_dialog_response (GtkWidget       *dialog,
+crop_dialog_response (CcAvatarChooser *self,
                       gint             response_id,
-                      CcAvatarChooser *self)
+                      GtkWidget       *dialog)
 {
-        GdkPixbuf *pb, *pb2;
+        g_autoptr(GdkPixbuf) pb = NULL;
+        g_autoptr(GdkPixbuf) pb2 = NULL;
 
         if (response_id != GTK_RESPONSE_ACCEPT) {
                 self->crop_area = NULL;
@@ -86,9 +87,6 @@ crop_dialog_response (GtkWidget       *dialog,
 
         set_user_icon_data (self->user, pb2);
 
-        g_object_unref (pb2);
-        g_object_unref (pb);
-
         self->crop_area = NULL;
         gtk_widget_destroy (dialog);
 
@@ -113,8 +111,8 @@ cc_avatar_chooser_crop (CcAvatarChooser *self,
 
         gtk_window_set_icon_name (GTK_WINDOW (dialog), "system-users");
 
-        g_signal_connect (G_OBJECT (dialog), "response",
-                          G_CALLBACK (crop_dialog_response), self);
+        g_signal_connect_object (G_OBJECT (dialog), "response",
+                                 G_CALLBACK (crop_dialog_response), self, G_CONNECT_SWAPPED);
 
         /* Content */
         self->crop_area = cc_crop_area_new ();
@@ -132,51 +130,52 @@ cc_avatar_chooser_crop (CcAvatarChooser *self,
 }
 
 static void
-file_chooser_response (GtkDialog       *chooser,
+file_chooser_response (CcAvatarChooser *self,
                        gint             response,
-                       CcAvatarChooser *self)
+                       GtkDialog       *chooser)
 {
-        gchar *filename;
-        GError *error;
-        GdkPixbuf *pixbuf, *pixbuf2;
+        g_autofree gchar *fileuri = NULL;
+        g_autoptr(GError) error = NULL;
+        g_autoptr(GdkPixbuf) pixbuf = NULL;
+        g_autoptr(GdkPixbuf) pixbuf2 = NULL;
+        g_autoptr(GFile) file = NULL;
+        g_autoptr(GFileInputStream) stream = NULL;
 
         if (response != GTK_RESPONSE_ACCEPT) {
                 gtk_widget_destroy (GTK_WIDGET (chooser));
                 return;
         }
 
-        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
+        fileuri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (chooser));
+        file = g_file_new_for_uri (fileuri);
 
-        error = NULL;
-        pixbuf = gdk_pixbuf_new_from_file (filename, &error);
+        stream = g_file_read (file, NULL, &error);
+        pixbuf = gdk_pixbuf_new_from_stream (G_INPUT_STREAM (stream),
+                                             NULL, &error);
         if (pixbuf == NULL) {
-                g_warning ("Failed to load %s: %s", filename, error->message);
-                g_error_free (error);
+                g_warning ("Failed to load %s: %s", fileuri, error->message);
         }
-        g_free (filename);
 
         pixbuf2 = gdk_pixbuf_apply_embedded_orientation (pixbuf);
-        g_object_unref (pixbuf);
 
         gtk_widget_destroy (GTK_WIDGET (chooser));
 
         cc_avatar_chooser_crop (self, pixbuf2);
-        g_object_unref (pixbuf2);
 }
 
 static void
 update_preview (GtkFileChooser               *chooser,
                 GnomeDesktopThumbnailFactory *thumb_factory)
 {
-        gchar *uri;
+        g_autofree gchar *uri = NULL;
 
         uri = gtk_file_chooser_get_uri (chooser);
 
         if (uri) {
-                GdkPixbuf *pixbuf = NULL;
-                char *mime_type = NULL;
-                GFile *file;
-                GFileInfo *file_info;
+                g_autoptr(GdkPixbuf) pixbuf = NULL;
+                g_autofree char *mime_type = NULL;
+                g_autoptr(GFile) file = NULL;
+                g_autoptr(GFileInfo) file_info = NULL;
                 GtkWidget *preview;
 
                 preview = gtk_file_chooser_get_preview_widget (chooser);
@@ -186,19 +185,16 @@ update_preview (GtkFileChooser               *chooser,
                                                "standard::*",
                                                G_FILE_QUERY_INFO_NONE,
                                                NULL, NULL);
-                g_object_unref (file);
 
                 if (file_info != NULL &&
                     g_file_info_get_file_type (file_info) != G_FILE_TYPE_DIRECTORY) {
                         mime_type = g_strdup (g_file_info_get_content_type (file_info));
-                        g_object_unref (file_info);
                 }
 
                 if (mime_type) {
                         pixbuf = gnome_desktop_thumbnail_factory_generate_thumbnail (thumb_factory,
                                                                                      uri,
                                                                                      mime_type);
-                        g_free (mime_type);
                 }
 
                 gtk_dialog_set_response_sensitive (GTK_DIALOG (chooser),
@@ -207,15 +203,12 @@ update_preview (GtkFileChooser               *chooser,
 
                 if (pixbuf != NULL) {
                         gtk_image_set_from_pixbuf (GTK_IMAGE (preview), pixbuf);
-                        g_object_unref (pixbuf);
                 }
                 else {
                         gtk_image_set_from_icon_name (GTK_IMAGE (preview),
                                                       "dialog-question",
                                                       GTK_ICON_SIZE_DIALOG);
                 }
-
-                g_free (uri);
         }
 
         gtk_file_chooser_set_preview_widget_active (chooser, TRUE);
@@ -242,6 +235,7 @@ cc_avatar_chooser_select_file (CcAvatarChooser *self)
         gtk_widget_set_size_request (preview, 128, -1);
         gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (chooser), preview);
         gtk_file_chooser_set_use_preview_label (GTK_FILE_CHOOSER (chooser), FALSE);
+        gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (chooser), FALSE);
         gtk_widget_show (preview);
 
         /* Preview has to be generated after default handler of "selection-changed"
@@ -260,8 +254,8 @@ cc_avatar_chooser_select_file (CcAvatarChooser *self)
         gtk_file_filter_add_pixbuf_formats (filter);
         gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (chooser), filter);
 
-        g_signal_connect (chooser, "response",
-                          G_CALLBACK (file_chooser_response), self);
+        g_signal_connect_object (chooser, "response",
+                                 G_CALLBACK (file_chooser_response), self, G_CONNECT_SWAPPED);
 
         gtk_window_present (GTK_WINDOW (chooser));
 }
@@ -275,20 +269,18 @@ destroy_chooser (GtkWidget *chooser)
 }
 
 static void
-webcam_response_cb (GtkDialog        *dialog,
+webcam_response_cb (CcAvatarChooser  *self,
                     int               response,
-                    CcAvatarChooser  *self)
+                    GtkDialog        *dialog)
 {
         if (response == GTK_RESPONSE_ACCEPT) {
-                GdkPixbuf *pb, *pb2;
+                g_autoptr(GdkPixbuf) pb = NULL;
+                g_autoptr(GdkPixbuf) pb2 = NULL;
 
                 g_object_get (G_OBJECT (dialog), "pixbuf", &pb, NULL);
                 pb2 = gdk_pixbuf_scale_simple (pb, PIXEL_SIZE, PIXEL_SIZE, GDK_INTERP_BILINEAR);
 
                 set_user_icon_data (self->user, pb2);
-
-                g_object_unref (pb2);
-                g_object_unref (pb);
         }
         if (response != GTK_RESPONSE_DELETE_EVENT &&
             response != GTK_RESPONSE_NONE)
@@ -319,8 +311,8 @@ webcam_icon_selected (CcAvatarChooser *self)
                     G_CALLBACK(on_focus_out_event),
                     GTK_WINDOW (window));
         gtk_window_set_modal (GTK_WINDOW (window), TRUE);
-        g_signal_connect (G_OBJECT (window), "response",
-                          G_CALLBACK (webcam_response_cb), self);
+        g_signal_connect_object (G_OBJECT (window), "response",
+                                 G_CALLBACK (webcam_response_cb), self, G_CONNECT_SWAPPED);
         gtk_widget_show (window);
 }
 
@@ -334,18 +326,14 @@ update_photo_menu_status (CcAvatarChooser *self)
 }
 
 static void
-device_added (CheeseCameraDeviceMonitor   *monitor,
-              CheeseCameraDevice          *device,
-              CcAvatarChooser             *self)
+device_added (CcAvatarChooser *self)
 {
         self->num_cameras++;
         update_photo_menu_status (self);
 }
 
 static void
-device_removed (CheeseCameraDeviceMonitor   *monitor,
-                const char                  *id,
-                CcAvatarChooser             *self)
+device_removed (CcAvatarChooser *self)
 {
         self->num_cameras--;
         update_photo_menu_status (self);
@@ -354,9 +342,8 @@ device_removed (CheeseCameraDeviceMonitor   *monitor,
 #endif /* HAVE_CHEESE */
 
 static void
-face_widget_activated (GtkFlowBox        *flowbox,
-                       GtkFlowBoxChild   *child,
-                       CcAvatarChooser   *self)
+face_widget_activated (CcAvatarChooser *self,
+                       GtkFlowBoxChild *child)
 {
         const gchar *filename;
         GtkWidget   *image;
@@ -402,9 +389,10 @@ create_face_widget (gpointer item,
 static void
 setup_cheese_camera_device_monitor (CcAvatarChooser *self)
 {
-        g_signal_connect (G_OBJECT (self->monitor), "added", G_CALLBACK (device_added), self);
-        g_signal_connect (G_OBJECT (self->monitor), "removed", G_CALLBACK (device_removed), self);
+        g_signal_connect_object (G_OBJECT (self->monitor), "added", G_CALLBACK (device_added), self, G_CONNECT_SWAPPED);
+        g_signal_connect_object (G_OBJECT (self->monitor), "removed", G_CALLBACK (device_removed), self, G_CONNECT_SWAPPED);
         cheese_camera_device_monitor_coldplug (self->monitor);
+        update_photo_menu_status (self);
 }
 
 static void
@@ -424,35 +412,58 @@ cheese_camera_device_monitor_new_cb (GObject *source,
 }
 #endif /* HAVE_CHEESE */
 
-static void
-setup_photo_popup (CcAvatarChooser *self)
+static GStrv
+get_settings_facesdirs (void)
 {
-        GFile *file, *dir;
-        GFileInfo *info;
-        GFileEnumerator *enumerator;
+        g_autoptr(GSettings) settings = g_settings_new ("org.gnome.desktop.interface");
+        g_auto(GStrv) settings_dirs = g_settings_get_strv (settings, "avatar-directories");
+        GPtrArray *facesdirs = g_ptr_array_new ();
+
+        if (settings_dirs != NULL) {
+                int i;
+                for (i = 0; settings_dirs[i] != NULL; i++) {
+                        char *path = settings_dirs[i];
+                        if (g_strcmp0 (path, "") != 0)
+                                g_ptr_array_add (facesdirs, g_strdup (path));
+                }
+        }
+        g_ptr_array_add (facesdirs, NULL);
+
+        return (GStrv) g_ptr_array_steal (facesdirs, NULL);
+}
+
+static GStrv
+get_system_facesdirs (void)
+{
+        const char * const * data_dirs;
+        GPtrArray *facesdirs;
+        int i;
+
+        facesdirs = g_ptr_array_new ();
+
+        data_dirs = g_get_system_data_dirs ();
+        for (i = 0; data_dirs[i] != NULL; i++) {
+                char *path = g_build_filename (data_dirs[i], "pixmaps", "faces", NULL);
+                g_ptr_array_add (facesdirs, path);
+        }
+        g_ptr_array_add (facesdirs, NULL);
+        return (GStrv) g_ptr_array_steal (facesdirs, NULL);
+}
+
+static gboolean
+add_faces_from_dirs (GListStore *faces, GStrv facesdirs, gboolean add_all)
+{
+        GFile *file;
         GFileType type;
         const gchar *target;
-        const gchar * const * dirs;
         guint i;
-        gboolean added_faces;
+        gboolean added_faces = FALSE;
 
-        self->faces = g_list_store_new (G_TYPE_FILE);
-        gtk_flow_box_bind_model (GTK_FLOW_BOX (self->flowbox),
-                                 G_LIST_MODEL (self->faces),
-                                 create_face_widget,
-                                 self,
-                                 NULL);
+        for (i = 0; facesdirs[i] != NULL; i++) {
+                g_autoptr(GFile) dir = NULL;
+                g_autoptr(GFileEnumerator) enumerator = NULL;
 
-        g_signal_connect (self->flowbox, "child-activated",
-                          G_CALLBACK (face_widget_activated), self);
-
-        dirs = g_get_system_data_dirs ();
-        for (i = 0; dirs[i] != NULL; i++) {
-                char *path;
-
-                path = g_build_filename (dirs[i], "pixmaps", "faces", NULL);
-                dir = g_file_new_for_path (path);
-                g_free (path);
+                dir = g_file_new_for_path (facesdirs[i]);
 
                 enumerator = g_file_enumerate_children (dir,
                                                         G_FILE_ATTRIBUTE_STANDARD_NAME ","
@@ -462,39 +473,62 @@ setup_photo_popup (CcAvatarChooser *self)
                                                         G_FILE_QUERY_INFO_NONE,
                                                         NULL, NULL);
                 if (enumerator == NULL) {
-                        g_object_unref (dir);
                         continue;
                 }
 
-                while ((info = g_file_enumerator_next_file (enumerator, NULL, NULL)) != NULL) {
-                        added_faces = TRUE;
+                while (TRUE) {
+                        g_autoptr(GFileInfo) info = g_file_enumerator_next_file (enumerator, NULL, NULL);
+                        if (info == NULL) {
+                                break;
+                        }
 
                         type = g_file_info_get_file_type (info);
                         if (type != G_FILE_TYPE_REGULAR &&
                             type != G_FILE_TYPE_SYMBOLIC_LINK) {
-                                g_object_unref (info);
                                 continue;
                         }
 
                         target = g_file_info_get_symlink_target (info);
                         if (target != NULL && g_str_has_prefix (target , "legacy/")) {
-                                g_object_unref (info);
                                 continue;
                         }
 
                         file = g_file_get_child (dir, g_file_info_get_name (info));
-                        g_list_store_append (self->faces, file);
+                        g_list_store_append (faces, file);
 
-                        g_object_unref (info);
+                        added_faces = TRUE;
                 }
 
                 g_file_enumerator_close (enumerator, NULL, NULL);
-                g_object_unref (enumerator);
-                g_object_unref (dir);
 
-                if (added_faces)
+                if (added_faces && !add_all)
                         break;
         }
+        return added_faces;
+}
+
+
+static void
+setup_photo_popup (CcAvatarChooser *self)
+{
+        g_auto(GStrv) settings_facesdirs = NULL;
+
+        self->faces = g_list_store_new (G_TYPE_FILE);
+        gtk_flow_box_bind_model (GTK_FLOW_BOX (self->flowbox),
+                                 G_LIST_MODEL (self->faces),
+                                 create_face_widget,
+                                 self,
+                                 NULL);
+
+        g_signal_connect_object (self->flowbox, "child-activated",
+                                 G_CALLBACK (face_widget_activated), self, G_CONNECT_SWAPPED);
+
+        settings_facesdirs = get_settings_facesdirs ();
+
+        if (!add_faces_from_dirs (self->faces, settings_facesdirs, TRUE)) {
+                g_auto(GStrv) system_facesdirs = get_system_facesdirs ();
+                add_faces_from_dirs (self->faces, system_facesdirs, FALSE);
+        }
 
 #ifdef HAVE_CHEESE
         gtk_widget_set_visible (self->take_picture_button, TRUE);
@@ -510,24 +544,22 @@ setup_photo_popup (CcAvatarChooser *self)
 }
 
 static void
-popup_icon_menu (GtkToggleButton *button,
-                 CcAvatarChooser *self)
+popup_icon_menu (CcAvatarChooser *self)
 {
         gtk_popover_popup (GTK_POPOVER (self));
 }
 
 static gboolean
-on_popup_button_button_pressed (GtkToggleButton *button,
-                                GdkEventButton  *event,
-                                CcAvatarChooser *self)
+on_popup_button_button_pressed (CcAvatarChooser *self,
+                                GdkEventButton  *event)
 {
         if (event->button == 1) {
                 if (!gtk_widget_get_visible (GTK_WIDGET (self))) {
-                        popup_icon_menu (button, self);
-                        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
+                        popup_icon_menu (self);
+                        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->popup_button), TRUE);
                 } else {
                         gtk_popover_popdown (GTK_POPOVER (self));
-                        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), FALSE);
+                        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->popup_button), FALSE);
                 }
 
                 return TRUE;
@@ -550,10 +582,10 @@ cc_avatar_chooser_new (GtkWidget *button)
         /* Set up the popup */
         self->popup_button = button;
         setup_photo_popup (self);
-        g_signal_connect (button, "toggled",
-                          G_CALLBACK (popup_icon_menu), self);
-        g_signal_connect (button, "button-press-event",
-                          G_CALLBACK (on_popup_button_button_pressed), self);
+        g_signal_connect_object (button, "toggled",
+                                 G_CALLBACK (popup_icon_menu), self, G_CONNECT_SWAPPED);
+        g_signal_connect_object (button, "button-press-event",
+                                 G_CALLBACK (on_popup_button_button_pressed), self, G_CONNECT_SWAPPED);
 
         return self;
 }
@@ -601,9 +633,7 @@ cc_avatar_chooser_class_init (CcAvatarChooserClass *klass)
 }
 
 static void
-user_flowbox_activated (GtkFlowBox        *flowbox,
-                        GtkFlowBoxChild   *child,
-                        CcAvatarChooser   *self)
+user_flowbox_activated (CcAvatarChooser *self)
 {
         set_default_avatar (self->user);
 
@@ -622,8 +652,7 @@ cc_avatar_chooser_set_user (CcAvatarChooser *self,
 
         if (self->user) {
                 gtk_container_foreach (GTK_CONTAINER (self->user_flowbox), (GtkCallback) gtk_widget_destroy, NULL);
-                g_object_unref (self->user);
-                self->user = NULL;
+                g_clear_object (&self->user);
         }
         self->user = g_object_ref (user);
 
@@ -633,6 +662,6 @@ cc_avatar_chooser_set_user (CcAvatarChooser *self,
         gtk_image_set_pixel_size (GTK_IMAGE (image), AVATAR_CHOOSER_PIXEL_SIZE);
         gtk_widget_show (image);
         gtk_container_add (GTK_CONTAINER (self->user_flowbox), image);
-        g_signal_connect (self->user_flowbox, "child-activated", G_CALLBACK (user_flowbox_activated), self);
+        g_signal_connect_object (self->user_flowbox, "child-activated", G_CALLBACK (user_flowbox_activated), self, G_CONNECT_SWAPPED);
 }
 
